"use strict";

/* globals process, require, module */

var meta = module.parent.require('./meta');
var user = module.parent.require('./user');
var SocketPlugins = require.main.require('./src/socket.io/plugins');

var _ = module.parent.require('underscore');
var winston = module.parent.require('winston');
var async = require('async');
var db = module.parent.require('./database');
var nconf = module.parent.require('nconf');
var groups = module.parent.require('./groups');
var Plugins = module.parent.require('./plugins');

var jwt = require('jsonwebtoken');
var fs = require('fs');
var co = require('co');
var dateFormat = require('dateformat');

var OSS = require('ali-oss');
var AV = require('leanengine');

var controllers = require('./lib/controllers');
var nbbAuthController = module.parent.require('./controllers/authentication');

/* Websocket Listeners */

SocketPlugins.ginseng = {};

SocketPlugins.ginseng.showUserIds = function(socket, data, callback) {
	// Retrieve the hash and find matches
	var uids = data.uids;
	var payload = [];
	var match, idx;

	payload.length = uids.length;

	if (uids.length) {
		db.getObject(plugin.settings.name + ':uid', function(err, hash) {
			for(var remoteId in hash) {
				idx = uids.indexOf(hash[remoteId]);
				if (hash.hasOwnProperty(remoteId) && idx !== -1) {
					payload[idx] = remoteId;
				}
			}

			callback(null, payload);
		});
	} else {
		callback(new Error('no-uids-supplied'));
	}
};

SocketPlugins.ginseng.findUserByRemoteId = function(socket, data, callback) {
	if (data.remoteId) {
		plugin.getUser(data.remoteId, callback);
	} else {
		callback(new Error('no-remote-id-supplied'));
	}
};

/* End Websocket Listeners */

var plugin = {
		ready: false,
		aliOSSClient: null,
		settings: {
			name: 'app',
			cookieName: 'token',
			cookieDomain: undefined,
			secret: '',
			behaviour: 'trust',
			noRegistration: 'off',
			OSSProvinder: '',
			aliOSSRegion: '',
			aliOSSKeyId: '',
			aliOSSKeySecret: '',
			aliOSSBucket: '',
			lcAppId: '',
			lcAppKey: '',
			lcAppMasterKey: '',
		} 
	};

plugin.init = function(params, callback) {
	var router = params.router;
	var hostMiddleware = params.middleware;

	router.get('/admin/plugins/ginseng', hostMiddleware.admin.buildHeader, controllers.renderAdminPage);
	router.get('/api/admin/plugins/ginseng', controllers.renderAdminPage);
	router.get('/api/ginseng/lookup', controllers.retrieveUser);

	if (process.env.NODE_ENV === 'development') {
		router.get('/debug/session', plugin.generate);
	}

	plugin.reloadSettings((err)=>{
		if (err) {
			return callback(err);
		}

		winston.warn('[GINSENG] settings:' + JSON.stringify(plugin.settings));

		if (plugin.settings.OSSProvinder == 'AliYun') {
			if (plugin.settings.aliOSSEndPoint && 
				plugin.settings.aliOSSKeyId && 
				plugin.settings.aliOSSKeySecret && 
				plugin.settings.aliOSSBucket) {
				plugin.aliOSSClient = new OSS({
				  accessKeyId: plugin.settings.aliOSSKeyId,
				  accessKeySecret: plugin.settings.aliOSSKeySecret,
				  endpoint: plugin.settings.aliOSSEndPoint,
				  bucket: plugin.settings.aliOSSBucket
				});				
				Plugins.registerHook("nodebb-plugin-ginseng", {hook: "filter:uploadImage", method: plugin.uploadImage_ali});
				Plugins.registerHook("nodebb-plugin-ginseng", {hook: "filter:uploadFile", method: plugin.uploadFile_ali});
				return callback();
			}
		} else if (plugin.settings.OSSProvinder == 'LeanCloud') {
		 	if (plugin.settings.lcAppId && plugin.settings.lcAppKey) {
				AV.init({
				  appId: plugin.settings.lcAppId,
				  appKey: plugin.settings.lcAppKey,
				  masterKey: plugin.lcAppMasterKey
				});
				if (plugin.settings.lcAppMasterKey) {
					AV.Cloud.useMasterKey();
				}
				Plugins.registerHook("nodebb-plugin-ginseng", {hook: "filter:uploadImage", method: plugin.uploadImage_leancloud});
				Plugins.registerHook("nodebb-plugin-ginseng", {hook: "filter:uploadFile", method: plugin.uploadFile_leancloud});
				return callback();
			}	
		}
		winston.warn('[GINSENG] invalid setting of OSS provinder [' + plugin.settings.OSSProvinder + '], file uploading disabled.');
		Plugins.registerHook("nodebb-plugin-ginseng", {hook: "filter:uploadImage", method: plugin.uploadImage_disabled});
		Plugins.registerHook("nodebb-plugin-ginseng", {hook: "filter:uploadFile", method: plugin.uploadFile_disabled});
		return callback();
	});
};

plugin.appendConfig = function(config, callback) {
	config.ginseng = {
		logoutRedirect: plugin.settings.logoutRedirect,
		loginOverride: plugin.settings.loginOverride
	};

	callback(null, config);
};


/*
 *	Given a remoteId, show user data
 */
plugin.getUser = function(remoteId, callback) {
	async.waterfall([
		async.apply(db.getObjectField, plugin.settings.name + ':uid', remoteId),
		function(uid, next) {
			if (uid) {
				user.getUserFields(uid, ['username', 'userslug', 'picture'], next);
			} else {
				setImmediate(next);
			}
		}
	], callback);
};

plugin.process = function(token, callback) {
	async.waterfall([
		async.apply(jwt.verify, token, plugin.settings.secret),
		async.apply(plugin.verifyToken),
		async.apply(plugin.findUser),
		async.apply(plugin.verifyUser)
	], callback);
};

plugin.verifyToken = function(payload, callback) {
	var originid =  payload['originid'];
	var role =  payload['role'];
	var username =payload['username'];

	winston.verbose(`[GINSENG] originid:{originid} role:{role} username:{username}`)
	if (!originid || !role || (!username)) {
		return callback(new Error('payload-invalid'));
	}

	callback(null, payload);
};

plugin.verifyUser = function(uid, callback) {
	// Check ban state of user, reject if banned
	user.isBanned(uid, function(err, banned) {
		callback(err || banned ? new Error('banned') : null, uid);
	});
};

plugin.findUser = function(payload, callback) {
	// If payload id resolves to a user, return the uid, otherwise register a new user
	winston.verbose('[GINSENG] Payload verified');

	var originid = payload['originid'];
	var email = payload['email'];
	var username = payload['username'];

	async.parallel({
		uid: async.apply(db.getObjectField, plugin.settings.name + ':uid', originid),
		mergeUid: async.apply(db.sortedSetScore, 'email:uid', email)
	}, function(err, checks) {
		if (err) { return callback(err); }

		if (checks.uid && !isNaN(parseInt(checks.uid, 10))) {
			// Ensure the uid exist
			user.exists(parseInt(checks.uid, 10), function(err, exists) {
				if (err) {
					return callback(err);
				} else if (exists) {
					return callback(null, checks.uid);
				} else {
					async.series([
						async.apply(db.deleteObjectField, plugin.settings.name + ':uid', originid),	// reference is outdated, user got deleted
						async.apply(plugin.createUser, payload)
					], function(err, data) {
						callback(err, data[1]);
					});
				}
			});
		} else if (email && email.length && checks.mergeUid && !isNaN(parseInt(checks.mergeUid, 10))) {
			winston.info('[GINSENG] Found user via their email, associating this id (' + originid + ') with their NodeBB account');
			db.setObjectField(plugin.settings.name + ':uid', originid, checks.mergeUid, function (err) {
				callback(err, checks.mergeUid);
			});
		} else {
			// No match, create a new user
			plugin.createUser(payload, callback);
		}
	});
};

plugin.createUser = function(payload, callback) {
	if (plugin.settings.noRegistration === 'on') {
		return callback(new Error('no-match'));
	}

	var originid = payload['originid'];
	var username = payload['username'];
	var email = payload['email'];
	var role = payload['role'];

	winston.info('[GINSENG] No user found, creating a new user for this login');
	username = username.trim().replace(/[^'"\s\-.*0-9\u00BF-\u1FFF\u2C00-\uD7FF\w]+/, '-');

	user.create({
		username: username,
		email: email
	}, function(err, uid) {
		if (err) { return callback(err); }

		delete payload['username'];
		delete payload['email'];

		Object.keys(payload).forEach((k, idx)=>{
			user.setUserField(uid, k, payload[k]);
		})
		
		groups.join('#'+role, uid);

		db.setObjectField(plugin.settings.name + ':uid', originid, uid, function (err) {
			callback(err, uid);
		});
	});
};


plugin.addMiddleware = function(req, res, next) {
	function handleGuest (req, res, next) {
		if (plugin.settings.guestRedirect && !req.originalUrl.startsWith(nconf.get('relative_path') + '/login?local=1')) {
			// If a guest redirect is specified, follow it
			res.redirect(plugin.settings.guestRedirect.replace('%1', encodeURIComponent(nconf.get('url') + req.originalUrl)));
		} else if (res.locals.fullRefresh === true) {
			res.redirect(req.url);
		} else {
			next();
		}
	}

	// Only respond to page loads by guests, not api or asset calls
	var hasSession = req.hasOwnProperty('user') && req.user.hasOwnProperty('uid') && parseInt(req.user.uid, 10) > 0;
	var hasLoginLock = req.session.hasOwnProperty('loginLock');

	if (
		!plugin.ready ||	// plugin not ready
		!plugin.settings.secret ||
		(plugin.settings.behaviour === 'trust' && hasSession) ||	// user logged in + "trust" behaviour
		(plugin.settings.behaviour === 'revalidate' && hasLoginLock) ||
		req.originalUrl.startsWith(nconf.get('relative_path') + '/api')	// api routes
	) {
		// Let requests through under "revalidate" behaviour only if they're logging in for the first time
		delete req.session.loginLock;	// remove login lock for "revalidate" logins
		return next();
	} else {
		// Hook into ip blacklist functionality in core
		if (meta.blacklist.test(req.ip)) {
			if (hasSession) {
				req.logout();
				res.locals.fullRefresh = true;
			}
			plugin.cleanup({ res: res });
			return handleGuest.apply(null, arguments);
		}

		if (Object.keys(req.cookies).length && req.cookies.hasOwnProperty(plugin.settings.cookieName) && req.cookies[plugin.settings.cookieName].length) {
			return plugin.process(req.cookies[plugin.settings.cookieName], function(err, uid) {
				if (err) {
					switch(err.message) {
						case 'banned':
							winston.info('[GINSENG] uid ' + uid + ' is banned, not logging them in');
							next();
							break;
						case 'payload-invalid':
							winston.warn('[GINSENG] The passed-in payload was invalid and could not be processed');
							next();
							break;
						case 'no-match':
							winston.info('[GINSENG] Payload valid, but local account not found.  Assuming guest.');
							handleGuest.call(null, req, res, next);
							break;
						default:
							winston.warn('[GINSENG] Error encountered while parsing token: ' + err.message);
							next();
							break;
					}
					return;
				}

				winston.info('[GINSENG] Processing login for uid ' + uid + ', path ' + req.originalUrl);
				req.uid = uid;
				nbbAuthController.doLogin(req, uid, function () {
					req.session.loginLock = true;
					res.redirect(req.originalUrl);
				});
			});
		} else if (hasSession) {
			// Has login session but no cookie, can assume "revalidate" behaviour
			user.isAdministrator(req.user.uid, function(err, isAdmin) {
				if (!isAdmin) {
					req.logout();
					res.locals.fullRefresh = true;
					handleGuest(req, res, next);
				} else {
					// Admins can bypass
					return next();
				}
			});
		} else {
			handleGuest.apply(null, arguments);
		}
	}
};

plugin.cleanup = function(data, callback) {
	if (plugin.settings.cookieDomain) {
		winston.verbose('[GINSENG] Clearing cookie');
		data.res.clearCookie(plugin.settings.cookieName, {
			domain: plugin.settings.cookieDomain,
			expires: new Date(),
			path: '/'
		});
	}

	if (typeof callback === 'function') {
		callback();
	} else {
		return true;
	}
};

plugin.generate = function(req, res) {
	var payload = {};
	payload[plugin.settings['payload:id']] = 1;
	payload[plugin.settings['payload:username']] = 'testUser';
	payload[plugin.settings['payload:email']] = 'testUser@example.org';
	payload[plugin.settings['payload:firstName']] = 'Test';
	payload[plugin.settings['payload:lastName']] = 'User';

	var token = jwt.sign(payload, plugin.settings.secret);
	res.cookie(plugin.settings.cookieName, token, {
		maxAge: 1000*60*60*24*21,
		httpOnly: true,
		domain: plugin.settings.cookieDomain
	});

	res.sendStatus(200);
};

plugin.addAdminNavigation = function(header, callback) {
	header.plugins.push({
		route: '/plugins/ginseng',
		icon: 'fa-user-secret',
		name: 'Ginseng'
	});

	callback(null, header);
};

plugin.reloadSettings = function(callback) {
	meta.settings.get('ginseng', function(err, settings) {
		if (err) {
			return callback(err);
		}

		winston.info('[GINSENG] Settings OK');
		plugin.settings = _.defaults(_.pick(settings, Boolean), plugin.settings);
		plugin.ready = true;

		callback();
	});
};


plugin.uploadImage_disabled = function(param, callback) {
	winston.info('[GINSENG] image uploading disabled:'+ JSON.stringify(param))
	callback({message:"[GINSENG] no OSS provinder, image uploading was disabled", param:param});
};

plugin.uploadFile_disables = function(param, callback) {
	winston.info('[GINSENG] file uploading disabled:'+ JSON.stringify(param))
	callback({message:"[GINSENG] no OSS provinder, file uploading was disabled", param:param});
};


plugin.uploadImage_leancloud = function(param, callback) {
	winston.info('[GINSENG] LeanCloud: image uploading:'+ JSON.stringify(param))
  	fs.readFile(param.image.path, function(err, data){
    	if (err) {
      		callback(err, param.image)
    	} else {
        	var theFile = new AV.File(param.image.name, data);
        	theFile.save().then(function(theFile){
        		param.image.url = theFile.url();
        		param.image.thumbnailURL = theFile.thumbnailURL(256, 256);
        		winston.info('save image success:'+ JSON.stringify(theFile))
        		winston.info('image url:'+ param.image.url)
        		winston.info('image thumbnail url:'+ param.image.thumbnailURL)
        		callback(null, param.image)
        	}, (err) => {
          		callback(err, param.image)
        	}).catch(callback);
        }
  	});
};

plugin.uploadFile_leancloud = function(param, callback) {
	winston.info('[GINSENG] LeanCloud: : file uploading:'+ JSON.stringify(param))

  	fs.readFile(param.file.path, function(err, data){
    	if (err) {
      		callback(err, param.file)
    	} else {
        	var theFile = new AV.File(param.file.name, data);
        	theFile.save().then(function(theFile){
        		param.file.url = theFile.url();
        		winston.info('save file success:'+ JSON.stringify(theFile))
        		winston.info('file url:'+ param.file.url)
        		callback(null, param.file)
        	}, (err) => {
          		callback(err, param.file)
        	}).catch(callback);
        }
  	});
};


plugin.makeFileObjectKey = function(folder, originFilename, uid) {
	let suffix = '';
	if (originFilename.lastIndexOf(".") != -1) {
		suffix = originFilename.substr(originFilename.lastIndexOf("."));
	}
	let now = new Date();
	return 'nodebb/'+folder+'/'+uid+'/'+now.getTime().toString(36)+suffix;
};

plugin.uploadImage_ali = function(param, callback) {
	winston.info('[GINSENG] Ali OSS: image uploading:'+ JSON.stringify(param))
	if (!plugin.aliOSSClient) {
		winston.error('[GINSENG] Ali OSS: client setup failed.');
		callback({message: "no-ali-oss-client"});
		return;
	}

	co(function*() {
		var result = yield plugin.aliOSSClient.put(plugin.makeFileObjectKey('image', param.image.name, param.uid), param.image.path, {uid: param.uid, origin:param});
		param.image.url = result.url;
	    winston.info('save image success:'+ JSON.stringify(result))
        winston.info('image url:'+ param.image.url)
	    callback(null, param.image)
	}).catch(callback);
};

plugin.uploadFile_ali = function(param, callback) {
	winston.info('[GINSENG] Ali OSS: file uploading:'+ JSON.stringify(param))
	if (!plugin.aliOSSClient) {
		winston.error('[GINSENG] Ali OSS: client setup failed.');
		callback({message: "no-ali-oss-client"});
		return;
	}

	co(function*() {
		var result = yield plugin.aliOSSClient.put(plugin.makeFileObjectKey('file', param.file.name, param.uid), param.file.path, {uid: param.uid, origin:param});
		param.file.url = result.url;
	    winston.info('save file success:'+ JSON.stringify(result))
        winston.info('file url:'+ param.file.url)
	    callback(null, param.file)
	}).catch(callback);
};




module.exports = plugin;